/*
 * Copyright (c) 2019, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#pragma once

#include <stdint.h>
#include <iostream>
#include <vector>
#include <utility>
#include <algorithm>
#include "common/cuml_comms_iface.hpp"


namespace MLCommon {
namespace Matrix {


/** Describes the data layout */
enum Layout {
  /** row major layout */
  LayoutRowMajor = 0,
  /** column major layout */
  LayoutColMajor
};


/**
 * @brief Descriptor data structure to provide info about the way data is
 *  distributed among the workers
 * <code>
 *   // single-node DGX2 with a fixed 2D block cyclic data distribution
 *   Descriptor md(40000, 40000, 1000, 1000, 4, 4);
 *   // print basic info
 *   printf("Total-size=%dx%d block-size=%dx%d\n", md.M, md.N, md.MB, md.NB);
 *   // print ownership of all blocks
 *   for(int i=0;i<md.numRowBlocks();++i) {
 *     for(int j=0;j<md.numColBlocks();++j) {
 *       printf("Device that owns global 2D-block row,col=%d,%d is %d\n",
 *              i, j, md.ranks[md.linearBlockIndex(i, j)]);
 *     }
 *   }
 *
 *   // single-node DGX1 with a custom data distribution
 *   Descriptor md1(40000, 40000, 20000, 10000, {0, 1, 2, 3, 4, 5, 6, 7});
 * </code>
 * @note Following assumptions come along with this structure:
 * <ul>
 *  <li>Block is a contiguous sub-matrix that's stored in one worker</li>
 *  <li>One worker can own multiple blocks</li>
 *  <li>Blocks are organized in column major order</li>
 * </ul>
 */
struct Descriptor {
  /** total number of rows */
  size_t M;
  /** total number of columns */
  size_t N;
  /** number of contiguous rows per block owned by a worker */
  size_t MB;
  /** number of contiguous cols per block owned by a worker */
  size_t NB;
  /** mapping of each block (in col-major order) to the device that owns it */
  std::vector<int> blocks2device;
  /**
   * intra block data layout. Note that inter-block layout is still assumed to
   * be column-major always!
   */
  Layout intraBlockLayout;
  /** mapping from global to local block (negative if block is not owned) */
  std::vector<int> global2local;

  /**
   * @brief For a given matrix and block-sizes construct the corresponding
   *  descriptor for it. This is useful when we are dealing with standard
   *  row/column-wise block-cyclic data distribution, as seen in other popular
   *  multi-node packages like magma etc.
   * @param _M total number of rows of this matrix
   * @param _N total number of columns
   * @param _MB number of rows in a block
   * @param _NB number of cols in a block
   * @param numDevicesRows number of devices across the rows
   * @param numDevicesCols number of devices across the cols
   * @param comm the communicator object
   * @param intraBlk layout inside of each block
   */
  Descriptor(size_t _M, size_t _N, size_t _MB, size_t _NB,
      size_t numDevicesRows, size_t numDevicesCols,
             const cumlCommunicator& comm, Layout intraBlk = LayoutRowMajor);

  /**
   * @brief For a given matrix and block-sizes construct the corresponding
   *  descriptor using the custom device allocation as specified by the user.
   * @param _M total number of rows of this matrix
   * @param _N total number of columns
   * @param _MB number of rows in a block
   * @param _NB number of cols in a block
   * @param b2d the custom blocks2devices assignment as done by the caller
   * @param comm the communicator object
   * @param intraBlk layout inside of each block
   */
  Descriptor(size_t _M, size_t _N, size_t _MB, size_t _NB,
             const std::vector<int> &b2d, const cumlCommunicator& comm,
             Layout intraBlk = LayoutRowMajor);

  /** total number of blocks across all workers */
  int totalBlocks() const { return numRowBlocks() * numColBlocks(); }

  /** total number of blocks along the row dimension */
  int numRowBlocks() const;

  /** total number of blocks along the column dimension */
  int numColBlocks() const;

  /**
   * @brief get the block's linear index from its global 2D-tileId
   * @param ridx row index of this block among all blocks
   * @param cidx col index of this block among all blocks
   * @return the linear index of this block in the `blocks2device` array
   */
  int linearBlockIndex(int ridx, int cidx) const;

  /**
   * @brief get the block's 2D-tileId (ridx,cidx) from its linear index
   * @param linear index of this block among all blocks
   * @return the pair (ridx,cidx) for the block specfied
   */
  std::pair<int,int> blockPosition(int linearBlockIndex) const;

  /** Count the total number of blocks owned by a given rank */
  int totalBlocksOwnedBy(int rank) const;

  /**
   * @brief Returns the vector of blocks (each identified by linearBLockIndex)
   * owned by the given rank
   */
  std::vector<int> blocksOwnedBy(int rank) const;

  /**
   * @brief Returns the 0-based offset of the block in the local buffer.
   * Increments by 1 for each local block
   */
  int blockOffset(int linearBlockIndex) const;

  /** Count the total number of matrix elements owned by a given rank */
  size_t totalElementsOwnedBy(int rank) const;

  friend std::ostream& operator<<(std::ostream& os, const Descriptor &desc);
  friend bool operator==(const Descriptor &a, const Descriptor &b);

  ///@todo: a converter to/from cudaLibDnDistMatrixDescriptor_t??
};


/** Print matrix descriptor in human readable form */
std::ostream& operator<<(std::ostream& os, const Descriptor &desc);

/** compare 2 descriptor objects */
bool operator==(const Descriptor &a, const Descriptor &b);

}; // end namespace Matrix
}; // end namespace MLCommon
